from abc import ABC
from typing import Any

from typing_extensions import override

from bolinette.core.exceptions import ExpressionError


class ExpressionNode(ABC):
    @override
    def __getattribute__(self, __name: str) -> "AttributeNode":
        return AttributeNode(self, __name)


class RootNode(ExpressionNode):
    pass


class AttributeNode(ExpressionNode):
    def __init__(self, parent: ExpressionNode, attr: str) -> None:
        self.parent = parent
        self.attr = attr


class ExpressionTree:
    def __init__(self) -> None:
        raise TypeError(f"Cannot instanciate {ExpressionTree}")

    @staticmethod
    def new() -> RootNode:
        return RootNode()

    @staticmethod
    def get_value(expr: ExpressionNode, obj: object) -> Any:
        cls = type(expr)
        if cls is RootNode:
            return obj
        if cls is AttributeNode:
            parent = object.__getattribute__(expr, "parent")
            attr = object.__getattribute__(expr, "attr")
            return getattr(ExpressionTree.get_value(parent, obj), attr)
        raise ExpressionError(f"Cannot get value from {type(expr).__name__}")

    @staticmethod
    def set_value(expr: ExpressionNode, obj: object, value: Any) -> None:
        cls = type(expr)
        if cls is AttributeNode:
            parent = object.__getattribute__(expr, "parent")
            attr = object.__getattribute__(expr, "attr")
            setattr(ExpressionTree.get_value(parent, obj), attr, value)
            return
        raise ExpressionError(f"Cannot set value of {type(expr).__name__}")

    @staticmethod
    def get_attribute_name(expr: AttributeNode) -> str:
        return object.__getattribute__(expr, "attr")



from collections.abc import Callable
from typing import Any, Awaitable, Literal, ParamSpec

from bolinette.core import Cache, __user_cache__, meta

P_Func = ParamSpec("P_Func")


class CommandMeta:
    def __init__(
        self,
        path: str,
        summary: str,
    ):
        self.path = path
        self.summary = summary


class Argument:
    def __init__(
        self,
        arg_type: Literal["argument", "option", "flag", "count"],
        name: str,
        *,
        flag: str | None = None,
        summary: str | None = None,
        value_type: type[Any] | None = None,
        default: Any | None = None,
        choices: list[str] | None = None,
    ):
        self.arg_type = arg_type
        self.name = name
        self.flag = flag
        self.summary = summary
        self.value_type = value_type
        self.default = default
        self.choices = choices


class ArgumentMeta(list[Argument]):
    pass


class _CommandDecorator:
    def __call__(self, name: str, summary: str, *, cache: Cache | None = None):
        def decorator(func: Callable[P_Func, Awaitable[None]]) -> Callable[P_Func, Awaitable[None]]:
            meta.set(func, CommandMeta(name, summary))
            (cache or __user_cache__).add(CommandMeta, func)
            return func

        return decorator

    def argument(
        self,
        arg_type: Literal["argument", "option", "flag", "count"],
        name: str,
        *,
        flag: str | None = None,
        summary: str | None = None,
        value_type: type | None = None,
        default: Any = None,
        choices: list[str] | None = None,
    ):
        def decorator(func: Callable[P_Func, Awaitable[None]]) -> Callable[P_Func, Awaitable[None]]:
            if not meta.has(func, ArgumentMeta):
                meta.set(func, ArgumentMeta())
            meta.get(func, ArgumentMeta).append(
                Argument(
                    arg_type,
                    name,
                    flag=flag,
                    summary=summary,
                    value_type=value_type,
                    default=default,
                    choices=choices,
                )
            )
            return func

        return decorator


command = _CommandDecorator()



from typing import Any, Callable, Generic, TypeVar

from typing_extensions import override

from bolinette.core.exceptions import MappingError
from bolinette.core.types import Type

SrcT = TypeVar("SrcT", bound=object)
DestT = TypeVar("DestT", bound=object)


class ForAttributeMapping:
    __slots__ = "dest_attr"

    def __init__(self, dest_attr: str) -> None:
        self.dest_attr = dest_attr


class MapFromAttribute(ForAttributeMapping):
    __slots__ = ("src_attr", "use_type")

    def __init__(self, src_attr: str, dest_attr: str) -> None:
        ForAttributeMapping.__init__(self, dest_attr)
        self.src_attr = src_attr
        self.use_type: Type[Any] | None = None


class IgnoreAttribute(ForAttributeMapping):
    __slots__ = "func"

    def __init__(self, dest_attr: str) -> None:
        ForAttributeMapping.__init__(self, dest_attr)


class MappingFunction(Generic[SrcT, DestT]):
    __slots__ = "func"

    def __init__(self, func: Callable[[SrcT, DestT], None]) -> None:
        self.func = func


class IncludeFromBase(Generic[SrcT, DestT]):
    __slots__ = ("src_t", "dest_t")

    def __init__(self, src_t: Type[SrcT], dest_t: Type[DestT]) -> None:
        self.src_t = src_t
        self.dest_t = dest_t


class MappingSequence(Generic[SrcT, DestT]):
    __slots__ = ("src_t", "dest_t", "head", "for_attrs", "tail", "includes")

    def __init__(self, src: type[SrcT], dest: type[DestT]) -> None:
        self.src_t = Type(src)
        self.dest_t = Type(dest)
        self.head: list[MappingFunction[SrcT, DestT]] = []
        self.for_attrs: dict[str, ForAttributeMapping] = {}
        self.tail: list[MappingFunction[SrcT, DestT]] = []
        self.includes: list[IncludeFromBase[Any, Any]] = []

    @staticmethod
    def get_hash(
        src_t: Type[Any],
        dest_t: Type[Any],
    ) -> int:
        return hash((src_t, dest_t))

    @override
    def __hash__(self) -> int:
        return MappingSequence.get_hash(self.src_t, self.dest_t)

    def add_head_func(self, func: MappingFunction[SrcT, DestT]) -> None:
        self.head.append(func)

    def add_for_attr(self, for_attr: ForAttributeMapping) -> None:
        self.for_attrs[for_attr.dest_attr] = for_attr

    def add_tail_func(self, func: MappingFunction[SrcT, DestT]) -> None:
        self.tail.append(func)

    def add_include(self, include: IncludeFromBase[Any, Any]) -> None:
        self.includes.append(include)

    def complete(self, completed: "dict[int, MappingSequence[Any, Any]]") -> None:
        head: list[MappingFunction[Any, Any]] = []
        for_attrs: dict[str, ForAttributeMapping] = {}
        tail: list[MappingFunction[Any, Any]] = []
        for include in self.includes:
            if (h := MappingSequence.get_hash(include.src_t, include.dest_t)) not in completed:
                raise MappingError(
                    f"Mapping ({self.src_t} -> {self.dest_t}): "
                    f"Could not find base mapping ({include.src_t} -> {include.dest_t}). "
                    "Make sure the mappings are declared in the right order."
                )
            base_seq = completed[h]
            head.extend(base_seq.head)
            for_attrs.update(base_seq.for_attrs)
            tail.extend(base_seq.tail)
        self.head = [*head, *self.head]
        self.for_attrs = {**for_attrs, **self.for_attrs}
        self.tail = [*tail, *self.tail]



from collections.abc import Iterator
from typing import Generic, Iterable, TypeVar

from typing_extensions import override

T = TypeVar("T")


class OrderedSet(Generic[T], set[T]):
    def __init__(self, iterable: Iterable[T] | None = None, /) -> None:
        super().__init__(iterable or ())
        self._order: list[T] = []

    @override
    def add(self, element: T, /) -> None:
        super().add(element)
        self._order.append(element)

    @override
    def remove(self, element: T, /) -> None:
        super().remove(element)
        self._order.remove(element)

    @override
    def pop(self) -> T:
        element = super().pop()
        self._order.remove(element)
        return element

    @override
    def discard(self, element: T, /) -> None:
        super().discard(element)
        if element in self._order:
            self._order.remove(element)

    @override
    def update(self, *sets: Iterable[T]) -> None:
        super().update(*sets)
        for s in sets:
            for element in s:
                if element not in self._order:
                    self._order.append(element)

    def __getitem__(self, key: int) -> T:
        return self._order[key]

    @override
    def __iter__(self) -> Iterator[T]:
        return iter(self._order)



import random
import re
import string


class StringUtils:
    __regex__ = {
        "snake": [
            re.compile(r"(.)([A-Z][a-z]+)"),
            re.compile(r"_+([A-Z])"),
            re.compile(r"([a-z0-9])([A-Z])"),
        ]
    }

    @staticmethod
    def to_snake_case(string: str):
        string = StringUtils.__regex__["snake"][0].sub(r"\1_\2", string)
        string = StringUtils.__regex__["snake"][1].sub(r"_\1", string)
        string = StringUtils.__regex__["snake"][2].sub(r"\1_\2", string)
        return string.lower()

    @staticmethod
    def capitalize(string: str):
        if not string:
            return string
        return f"{string[0].upper()}{string[1:]}"

    @staticmethod
    def random_string(length: int) -> str:
        return "".join(random.choices(string.ascii_letters + string.digits, k=length))



from typing import Any, Callable, Concatenate, ParamSpec, TypeVar, get_origin

from bolinette.core import Cache, __user_cache__, meta
from bolinette.core.injection.hook import InjectionHook
from bolinette.core.injection.registration import InjectionStrategy
from bolinette.core.types import Type

FuncP = ParamSpec("FuncP")
InstanceT = TypeVar("InstanceT")
InjectionSymbol = object()


class InitMethodMeta:
    pass


def init_method(func: Callable[Concatenate[InstanceT, FuncP], None]) -> Callable[Concatenate[InstanceT, FuncP], None]:
    meta.set(func, InitMethodMeta())
    return func


class InjectionParamsMeta:
    __slots__ = ("strategy", "args", "named_args", "before_init", "after_init", "match_all")

    def __init__(
        self,
        strategy: InjectionStrategy,
        args: list[Any] | None,
        named_args: dict[str, Any] | None,
        before_init: list[Callable[[InstanceT], None]] | None,
        after_init: list[Callable[[InstanceT], None]] | None,
        match_all: bool,
    ) -> None:
        self.strategy = strategy
        self.args = args or []
        self.named_args = named_args or {}
        self.before_init = before_init or []
        self.after_init = after_init or []
        self.match_all = match_all


def injectable(
    *,
    strategy: InjectionStrategy = "singleton",
    args: list[Any] | None = None,
    named_args: dict[str, Any] | None = None,
    cache: Cache | None = None,
    before_init: list[Callable[[InstanceT], None]] | None = None,
    after_init: list[Callable[[InstanceT], None]] | None = None,
    match_all: bool = False,
) -> Callable[[type[InstanceT]], type[InstanceT]]:
    def decorator(cls: type[InstanceT]) -> type[InstanceT]:
        if origin := get_origin(cls):
            _cls = origin
        else:
            _cls = cls
        meta.set(_cls, InjectionParamsMeta(strategy, args, named_args, before_init, after_init, match_all))
        (cache or __user_cache__).add(InjectionSymbol, cls)
        return cls

    return decorator


def require(cls: type[InstanceT]) -> Callable[[Callable[..., Any]], InjectionHook[InstanceT]]:
    def decorator(func: Callable[..., Any]) -> InjectionHook[InstanceT]:
        return InjectionHook(Type(cls))

    return decorator



from bolinette.core.version import __version__ as __version__
from bolinette.core.objects import GenericMeta as GenericMeta
from bolinette.core.metadata import meta as meta
from bolinette.core.cache import Cache as Cache, __user_cache__ as __user_cache__
from bolinette.core.logger import Logger as Logger, ConsoleColorCode as ConsoleColorCode
from bolinette.core.environment import (
    Environment as Environment,
    environment as environment,
    CoreSection as CoreSection,
)
from bolinette.core.command.command import command as command
from bolinette.core.extension import Extension as Extension, core_ext as core_ext
from bolinette.core.bolinette import Bolinette as Bolinette, main_func as main_func



from typing import Any, ParamSpec, TypeVar

P = ParamSpec("P")
InstanceT = TypeVar("InstanceT")


class Cache:
    def __init__(self, debug: bool = False) -> None:
        self.debug = debug
        self._bag: dict[Any, list[Any]] = {}

    def __contains__(self, key: Any) -> bool:
        return key in self._bag

    def get(self, key: Any, /, *, hint: type[InstanceT] | None = None, raises: bool = True) -> list[InstanceT]:
        if key not in self:
            if raises:
                raise KeyError(key)
            return []
        return self._bag[key]

    def delete(self, key: Any) -> None:
        if key not in self:
            raise KeyError(key)
        del self._bag[key]

    def init(self, key: Any) -> None:
        self._bag[key] = []

    def add(self, key: Any, value: Any) -> None:
        if key not in self:
            self.init(key)
        self._bag[key].append(value)

    def remove(self, key: Any, value: Any) -> None:
        if key not in self:
            raise KeyError(key)
        self._bag[key] = list(filter(lambda i: i is not value, self._bag[key]))

    @staticmethod
    def _merge(c1: "Cache", c2: "Cache") -> "Cache":
        c3 = Cache()
        for key, values in c1._bag.items():
            for value in values:
                c3.add(key, value)
        for key, values in c2._bag.items():
            for value in values:
                c3.add(key, value)
        return c3

    def __or__(self, __t: Any) -> "Cache":
        if isinstance(__t, Cache):
            return self._merge(self, __t)
        raise NotImplementedError()

    def __ror__(self, __t: Any) -> "Cache":
        if isinstance(__t, Cache):
            return self._merge(self, __t)
        raise NotImplementedError()


__user_cache__ = Cache()



import collections.abc
import contextlib
from types import NoneType, UnionType
from typing import (
    Any,
    ForwardRef,
    Generic,
    TypeVar,
    Union,
    get_args,
    get_origin,
    get_type_hints,
)

from typing_extensions import override

from bolinette.core import types
from bolinette.core.exceptions import TypingError

T = TypeVar("T")


class Type(Generic[T]):
    __slots__ = ("cls", "vars", "annotations", "nullable", "union")

    @staticmethod
    def from_instance(__instance: T) -> "Type[T]":
        return Type(type(__instance))

    def __init__(
        self,
        __cls: type[T],
        *,
        lookup: "types.TypeVarLookup[Any] | None" = None,
        raise_on_string: bool = True,
        raise_on_typevar: bool = True,
    ) -> None:
        if get_origin(__cls) in (Union, UnionType):
            args = get_args(__cls)
            self.nullable = None in args or NoneType in args
            args = tuple(a for a in args if a not in (None, NoneType))
            __cls, *additional_cls = args
        else:
            additional_cls = ()
            self.nullable = False
        self.cls, self.vars = Type.get_generics(__cls, lookup, raise_on_string, raise_on_typevar)
        self.vars = (*self.vars, *map(lambda _: Any, range(len(self.vars), self.get_param_count(self.cls))))
        self.annotations = Type._get_recursive_annotations(self.cls)
        self.union = tuple(Type(c) for c in additional_cls)

    @override
    def __str__(self) -> str:
        def _format_v(v: type[Any] | TypeVar | ForwardRef) -> str:
            if isinstance(v, type):
                return v.__qualname__
            if isinstance(v, TypeVar):
                return f"~{v.__name__}"
            if v is Ellipsis:
                return "..."
            return f"'{str(v.__forward_arg__)}'"

        if not self.vars:
            repr_str = _format_v(self.cls)
        else:
            repr_str = f"{_format_v(self.cls)}[{', '.join(map(_format_v, self.vars))}]"

        if self.is_union:
            for t in self.union:
                repr_str += f" | {str(t)}"

        if self.nullable:
            repr_str += " | None"

        return repr_str

    @override
    def __repr__(self) -> str:
        return f"<Type {str(self)}>"

    @override
    def __hash__(self) -> int:
        return hash((self.cls, self.vars))

    @override
    def __eq__(self, __value: object) -> bool:
        return (
            isinstance(__value, Type)
            and __value.cls is self.cls  # pyright: ignore[reportUnknownMemberType]
            and __value.vars == self.vars
        )

    def new(self, *args: Any, **kwargs: Any) -> T:
        return self.cls(*args, **kwargs)

    @property
    def is_union(self) -> bool:
        return len(self.union) > 0

    @property
    def is_any(self) -> bool:
        return self.cls is Any

    @staticmethod
    def _get_recursive_annotations(_cls: type[Any]) -> "dict[str, Type[Any]]":
        annotations: dict[str, Type[Any]] = {}
        try:
            for base in _cls.__bases__:
                annotations |= Type._get_recursive_annotations(base)
            hints: dict[str, type[Any]] = get_type_hints(_cls)
            for attr_name, hint in hints.items():
                annotations[attr_name] = Type(hint)
        except (AttributeError, TypeError, NameError):
            return annotations
        return annotations

    @staticmethod
    def get_generics(
        _cls: type[T],
        lookup: "types.TypeVarLookup[Any] | None",
        raise_on_string: bool,
        raise_on_typevar: bool,
    ) -> tuple[type[T], tuple[Any, ...]]:
        if origin := get_origin(_cls):
            type_vars: tuple[Any, ...] = ()
            for arg in get_args(_cls):
                if isinstance(arg, ForwardRef) and raise_on_string:
                    raise TypingError(
                        f"Generic parameter '{arg.__forward_arg__}' cannot be a string", cls=origin.__qualname__
                    )
                if isinstance(arg, TypeVar):
                    if lookup is not None:
                        if arg in lookup:
                            arg = lookup[arg]
                        elif raise_on_typevar:
                            raise TypingError(
                                f"TypeVar ~{arg.__name__} could not be found in lookup", cls=origin.__qualname__
                            )
                    elif raise_on_typevar:
                        raise TypingError(
                            f"Generic parameter ~{arg.__name__} cannot be a TypeVar", cls=origin.__qualname__
                        )
                type_vars = (*type_vars, arg)
            return origin, type_vars
        return _cls, ()

    @staticmethod
    def get_param_count(__cls: type[Any]) -> int:
        if __cls in _BUILTIN_PARAM_COUNT:
            return _BUILTIN_PARAM_COUNT[__cls]
        if hasattr(__cls, "__parameters__"):
            return len(__cls.__parameters__)  # type: ignore
        return 0


_BUILTIN_PARAM_COUNT: dict[type[Any], int] = {
    collections.abc.Hashable: 0,
    collections.abc.Awaitable: 1,
    collections.abc.Coroutine: 3,
    collections.abc.AsyncIterable: 1,
    collections.abc.AsyncIterator: 1,
    collections.abc.Iterable: 1,
    collections.abc.Iterator: 1,
    collections.abc.Reversible: 1,
    collections.abc.Sized: 0,
    collections.abc.Container: 1,
    collections.abc.Collection: 1,
    collections.abc.Set: 1,
    collections.abc.MutableSet: 1,
    collections.abc.Mapping: 2,
    collections.abc.MutableMapping: 2,
    collections.abc.Sequence: 1,
    collections.abc.MutableSequence: 1,
    collections.abc.ByteString: 0,
    list: 1,
    collections.deque: 1,
    set: 1,
    frozenset: 1,
    collections.abc.MappingView: 1,
    collections.abc.KeysView: 1,
    collections.abc.ItemsView: 2,
    collections.abc.ValuesView: 1,
    contextlib.AbstractContextManager: 1,
    contextlib.AbstractAsyncContextManager: 1,
    dict: 2,
    collections.defaultdict: 2,
    collections.OrderedDict: 2,
    collections.Counter: 1,
    collections.ChainMap: 2,
    collections.abc.Generator: 3,
    collections.abc.AsyncGenerator: 2,
    type: 1,
}



from typing import Any

import yaml

from bolinette.core.utils import paths


class FileUtils:
    def __init__(self, paths: paths.PathUtils) -> None:
        self._paths = paths

    @staticmethod
    def read_file(path: str) -> str:
        with open(path) as f:
            return f.read()

    @staticmethod
    def read_yaml(path: str) -> dict[str, Any]:
        with open(path) as f:
            return yaml.safe_load(f)

    def read_requirements(self, path: str, *, name: str = "requirements.txt") -> list[str]:
        return list(
            filter(
                lambda r: len(r),
                self.read_file(self._paths.join(path, name)).split("\n"),
            )
        )

    def read_profile(self, path: str) -> str | None:
        try:
            with open(self._paths.join(path, ".profile")) as f:
                for line in f:
                    return line
        except FileNotFoundError:
            return None
        return None



import os
from collections.abc import Callable
from typing import Any, Protocol, TypeVar

from bolinette.core import Cache, Logger, __user_cache__, meta
from bolinette.core.exceptions import EnvironmentError
from bolinette.core.injection import Injection, init_method
from bolinette.core.mapping import Mapper
from bolinette.core.utils import FileUtils, PathUtils


class _EnvSectionMeta:
    def __init__(self, name: str) -> None:
        self.name = name


class Environment:
    _OS_ENV_PREFIX = "BLNT_"

    def __init__(
        self,
        profile: str,
        logger: "Logger[Environment]",
        cache: Cache,
        inject: Injection,
        paths: PathUtils,
        files: FileUtils,
        mapper: Mapper,
        *,
        env_path: str = "env",
    ) -> None:
        self._profile = profile
        self._logger = logger
        self._cache = cache
        self._inject = inject
        self._paths = paths
        self._files = files
        self._mapper = mapper
        self._path = env_path
        self._raw_env: dict[str, Any] = {}

    def _init_section(self, section: object) -> None:
        section_name = meta.get(type(section), _EnvSectionMeta).name
        if section_name not in self._raw_env:
            raise EnvironmentError(f"No '{section_name}' section was found in the environment files")
        self._mapper.map(
            dict[str, Any],
            type(section),
            self._raw_env[section_name],
            section,
            src_path=f"Environment['{section_name}']",
        )

    @init_method
    def init(self) -> None:
        if EnvironmentSection in self._cache:
            for cls in self._cache.get(EnvironmentSection, hint=type):
                self._inject.add(cls, "singleton", before_init=[self._init_section])

        stack = [
            self._init_from_os(),
            self._init_from_file("env.yaml"),
            self._init_from_file(f"env.{self._profile}.yaml"),
            self._init_from_file(f"env.local.{self._profile}.yaml"),
        ]

        merged: dict[str, Any] = {}
        for node in stack:
            for name, section in node.items():
                if name not in merged:
                    merged[name] = {}
                for key, value in section.items():
                    merged[name][key] = value
        self._raw_env = merged

    @staticmethod
    def _init_from_os() -> dict[str, dict[str, Any]]:
        _vars: dict[str, str] = {}
        prefix_len = len(Environment._OS_ENV_PREFIX)
        for var in os.environ:
            if var.startswith(Environment._OS_ENV_PREFIX):
                _vars[var[prefix_len:]] = os.environ[var]
        _env: dict[str, Any] = {}
        for key, value in _vars.items():
            path = [s.lower() for s in key.split("__")]
            _node = _env
            for p in path[:-1]:
                if p not in _node:
                    _node[p] = {}
                _node = _node[p]
            if not isinstance(_node, dict) or path[-1] in _node:
                raise EnvironmentError(
                    f"OS variable '{Environment._OS_ENV_PREFIX}{key}' conflicts with other variables"
                )
            _node[path[-1]] = value
        return _env

    def _init_from_file(self, file_name: str) -> dict[str, dict[str, Any]]:
        try:
            return self._files.read_yaml(self._paths.env_path(file_name))
        except FileNotFoundError:
            return {}


class EnvironmentSection(Protocol):
    def __init__(self) -> None:
        pass


EnvT = TypeVar("EnvT", bound=EnvironmentSection)


def environment(name: str, *, cache: Cache | None = None) -> Callable[[type[EnvT]], type[EnvT]]:
    def decorator(cls: type[EnvT]) -> type[EnvT]:
        meta.set(cls, _EnvSectionMeta(name))
        (cache or __user_cache__).add(EnvironmentSection, cls)
        return cls

    return decorator


class CoreSection:
    debug: bool = False

    @init_method
    def _init_debug(self, cache: Cache) -> None:
        cache.debug = self.debug



from typing import Any, Callable, Protocol, TypeVar

from bolinette.core import Cache, __user_cache__, injection, meta
from bolinette.core.exceptions import InjectionError
from bolinette.core.injection.hook import InjectionHook
from bolinette.core.injection.registration import InjectionStrategy
from bolinette.core.types import Type
from bolinette.core.utils import OrderedSet


class ArgResolverOptions:
    __slots__ = (
        "injection",
        "caller",
        "caller_type_vars",
        "caller_strategy",
        "name",
        "t",
        "nullable",
        "default_set",
        "default",
        "immediate",
        "circular_guard",
    )

    def __init__(
        self,
        injection: "injection.Injection",
        caller: Callable[..., Any],
        caller_type_vars: tuple[Any, ...] | None,
        caller_strategy: InjectionStrategy,
        name: str,
        t: Type[Any],
        nullable: bool,
        default_set: bool,
        default: Any | None,
        immediate: bool,
        circular_guard: OrderedSet[Any],
    ) -> None:
        self.injection = injection
        self.caller = caller
        self.caller_type_vars = caller_type_vars
        self.caller_strategy = caller_strategy
        self.name = name
        self.t = t
        self.nullable = nullable
        self.default_set = default_set
        self.default = default
        self.immediate = immediate
        self.circular_guard = circular_guard


class ArgumentResolver(Protocol):
    def supports(self, options: ArgResolverOptions) -> bool:
        ...

    def resolve(self, options: ArgResolverOptions) -> tuple[str, Any]:
        ...


class ArgResolverMeta:
    __slots__ = ("priority", "scoped")

    def __init__(self, priority: int, scoped: bool) -> None:
        self.priority = priority
        self.scoped = scoped


ArgResolverT = TypeVar("ArgResolverT", bound=ArgumentResolver)


def injection_arg_resolver(
    *, priority: int = 0, scoped: bool = False, cache: Cache | None = None
) -> Callable[[type[ArgResolverT]], type[ArgResolverT]]:
    def decorator(cls: type[ArgResolverT]) -> type[ArgResolverT]:
        (cache or __user_cache__).add(ArgumentResolver, cls)
        meta.set(cls, ArgResolverMeta(priority, scoped))
        return cls

    return decorator


class DefaultArgResolver:
    __slots__ = ()

    def supports(self, options: "ArgResolverOptions") -> bool:
        return True

    def resolve(self, options: "ArgResolverOptions") -> tuple[str, Any]:
        if options.t.cls is Type:
            return (options.name, Type(options.t.vars[0]))
        if options.t.cls is type:
            return (options.name, options.t.vars[0])

        if not options.injection.is_registered(options.t):
            if options.nullable:
                return (options.name, None)
            if options.default_set:
                return (options.name, options.default)
            raise InjectionError(
                f"Type {options.t} is not a registered type in the injection system",
                func=options.caller,
                param=options.name,
            )

        r_type = options.injection.registered_types[options.t.cls].get_type(options.t)

        if r_type.strategy == "scoped":
            if options.caller_strategy in ["singleton", "transcient"]:
                raise InjectionError(
                    f"Cannot instanciate a scoped service in a {options.caller_strategy} service",
                    func=options.caller,
                    param=options.name,
                )

        if options.injection.__has_instance__(r_type):
            return (options.name, options.injection.__get_instance__(r_type))
        if options.immediate:
            return (
                options.name,
                options.injection.__instanciate__(r_type, options.t, options.circular_guard),
            )
        return (options.name, InjectionHook(options.t))



from bolinette.core.utils.paths import PathUtils as PathUtils
from bolinette.core.utils.files import FileUtils as FileUtils
from bolinette.core.utils.attrs import AttributeUtils as AttributeUtils
from bolinette.core.utils.strings import StringUtils as StringUtils
from bolinette.core.utils.collections import OrderedSet as OrderedSet



from typing import Callable, Protocol

from bolinette.core import (
    Environment,
    Extension,
    Logger,
    __user_cache__,
    core_ext,
    meta,
)
from bolinette.core.command import Parser
from bolinette.core.injection import Injection
from bolinette.core.utils import FileUtils, PathUtils


class _ExtensionModule(Protocol):
    __blnt_ext__: Extension


class Bolinette:
    def __init__(
        self,
        *,
        profile: str | None = None,
        inject: Injection | None = None,
        load_defaults: bool = True,
        extensions: list[_ExtensionModule] | None = None,
    ) -> None:
        if extensions is None:
            _extensions = []
        else:
            _extensions = [m.__blnt_ext__ for m in extensions]
        if load_defaults and core_ext not in _extensions:
            _extensions = [core_ext, *_extensions]

        _extensions = Extension.sort_extensions(_extensions)
        cache = Extension.merge_caches(_extensions)
        cache |= __user_cache__

        self._inject = inject or Injection(cache)
        meta.set(self, self._inject)

        self._logger = self._inject.require(Logger[Bolinette])
        self._paths = self._inject.require(PathUtils)
        self._files = self._inject.require(FileUtils)

        self._profile = profile or self._files.read_profile(self._paths.env_path()) or self._set_default_profile()

        self._inject.add(Bolinette, "singleton", instance=self)
        self._inject.add(
            Environment,
            "singleton",
            args=[self._profile],
            instanciate=True,
        )
        self._inject.__hook_proxies__(self)

    @property
    def injection(self) -> Injection:
        return self._inject

    def _set_default_profile(self) -> str:
        self._logger.warning(
            "No profile set, defaulting to 'development'.",
            "Be sure to set the current profile in a .profile file in the env folder",
        )
        return "development"

    async def startup(self) -> None:
        pass

    @property
    def _parser(self) -> Parser:
        ...

    async def exec_cmd_args(self):
        await self._parser.run()


def main_func(func: Callable[[], Bolinette]) -> Callable[[], Bolinette]:
    setattr(func, "__blnt__", "__blnt_main__")
    return func



import inspect
from collections.abc import Callable, Collection
from types import NoneType, UnionType
from typing import (
    TYPE_CHECKING,
    Any,
    Literal,
    ParamSpec,
    Protocol,
    TypeVar,
    Union,
    get_args,
    get_origin,
    get_type_hints,
    overload,
)

from typing_extensions import override

from bolinette.core import Cache, GenericMeta, meta
from bolinette.core.exceptions import InjectionError
from bolinette.core.injection.context import InjectionContext
from bolinette.core.injection.decorators import (
    InitMethodMeta,
    InjectionParamsMeta,
    InjectionSymbol,
)
from bolinette.core.injection.hook import InjectionHook, InjectionProxy
from bolinette.core.injection.registration import (
    InjectionStrategy,
    RegisteredType,
    RegisteredTypeBag,
)
from bolinette.core.injection.resolver import (
    ArgResolverMeta,
    ArgResolverOptions,
    ArgumentResolver,
    DefaultArgResolver,
)
from bolinette.core.types import Type, TypeVarLookup
from bolinette.core.utils import OrderedSet

FuncP = ParamSpec("FuncP")
FuncT = TypeVar("FuncT")
InstanceT = TypeVar("InstanceT")


class Injection:
    __slots__ = ("_cache", "_global_ctx", "_types", "_arg_resolvers")
    _ADD_INSTANCE_STRATEGIES = ("singleton",)
    _REQUIREABLE_STRATEGIES = ("singleton", "transcient")

    def __init__(
        self,
        cache: Cache,
        global_ctx: InjectionContext | None = None,
        types: "dict[type[Any], RegisteredTypeBag[Any]] | None" = None,
    ) -> None:
        self._arg_resolvers: list[ArgumentResolver] = []
        self._cache = cache
        self._global_ctx = global_ctx or InjectionContext()
        self._types = types if types is not None else self._pickup_types(cache)
        self._add_type_instance(Type(Cache), Type(Cache), False, "singleton", [], {}, [], [], cache, safe=True)
        self._add_type_instance(Type(Injection), Type(Injection), False, "singleton", [], {}, [], [], self, safe=True)
        self._arg_resolvers = self._pickup_resolvers(cache)

    @property
    def registered_types(self):
        return dict(self._types)

    @staticmethod
    def _pickup_types(cache: Cache) -> "dict[type[Any], RegisteredTypeBag[Any]]":
        if InjectionSymbol not in cache:
            return {}
        types: dict[type[Any], RegisteredTypeBag[Any]] = {}
        for cls in cache.get(InjectionSymbol, hint=type):
            t = Type(cls)
            if t.cls not in types:
                types[t.cls] = RegisteredTypeBag(t.cls)
            type_bag = types[t.cls]
            _meta = meta.get(t.cls, InjectionParamsMeta)
            if _meta.match_all:
                type_bag.set_match_all(
                    t,
                    _meta.strategy,  # type: ignore
                    _meta.args,
                    _meta.named_args,
                    _meta.before_init,
                    _meta.after_init,
                )
            else:
                type_bag.add_type(
                    t,
                    t,
                    _meta.strategy,  # type: ignore
                    _meta.args,
                    _meta.named_args,
                    _meta.before_init,
                    _meta.after_init,
                )
        return types

    def _pickup_resolvers(self, cache: Cache) -> "list[ArgumentResolver]":
        resolver_scoped: dict[type, bool] = {}
        resolver_priority: dict[type, int] = {}
        resolver_types: list[type[ArgumentResolver]] = []

        for t in cache.get(ArgumentResolver, hint=type[ArgumentResolver], raises=False):
            _meta = meta.get(t, ArgResolverMeta)
            resolver_priority[t] = _meta.priority
            resolver_scoped[t] = _meta.scoped
            resolver_types.append(t)
        resolver_types = sorted(resolver_types, key=lambda t: resolver_priority[t])

        resolvers: list[ArgumentResolver] = []
        for cls in resolver_types:
            if resolver_scoped[cls]:
                continue
            resolvers.append(self.instanciate(cls, additional_resolvers=[DefaultArgResolver()]))

        return [*resolvers, DefaultArgResolver()]

    def __has_instance__(self, r_type: "RegisteredType[Any]") -> bool:
        return r_type.strategy == "singleton" and self._global_ctx.has_instance(r_type.t)

    def __get_instance__(self, r_type: "RegisteredType[InstanceT]") -> InstanceT:
        return self._global_ctx.get_instance(r_type.t)

    def __set_instance__(self, r_type: "RegisteredType[InstanceT]", instance: InstanceT) -> None:
        if r_type.strategy == "singleton":
            self._global_ctx.set_instance(r_type.t, instance)

    def _resolve_args(
        self,
        func: Callable[..., Any],
        func_type_vars: tuple[Any, ...] | None,
        strategy: InjectionStrategy,
        vars_lookup: TypeVarLookup[Any] | None,
        immediate: bool,
        circular_guard: OrderedSet[Any],
        args: list[Any],
        named_args: dict[str, Any],
        additional_resolvers: list[ArgumentResolver],
    ) -> dict[str, Any]:
        if func in circular_guard:
            call_chain = " -> ".join(str(f) for f in circular_guard) + f" -> {func}"
            raise InjectionError(f"A circular call has been detected: {call_chain}", cls=circular_guard[0])
        circular_guard.add(func)

        func_params = dict(inspect.signature(func).parameters)
        if any((n, p) for n, p in func_params.items() if p.kind in (p.POSITIONAL_ONLY, p.VAR_POSITIONAL)):
            raise InjectionError(
                "Positional only parameters and positional wildcards are not allowed",
                func=func,
            )

        f_args: dict[str, Any] = {}
        _args = [*args]
        _named_args = {**named_args}

        try:
            if inspect.isclass(func):
                hints = get_type_hints(func.__init__)
            else:
                hints = get_type_hints(func)
        except NameError as exp:
            raise InjectionError(f"Type hint '{exp.name}' could not be resolved", func=func) from exp

        for p_name, param in func_params.items():
            if param.kind == param.VAR_KEYWORD:
                for kw_name, kw_value in _named_args.items():
                    f_args[kw_name] = kw_value
                _named_args = {}
                break

            if _args:
                f_args[p_name] = _args.pop(0)
                continue
            if p_name in _named_args:
                f_args[p_name] = _named_args.pop(p_name)
                continue

            default_set = False
            default = None
            if param.default is not param.empty:
                default_set = True
                default = param.default
            nullable = False

            if p_name not in hints:
                if default_set:
                    f_args[p_name] = default
                    continue
                raise InjectionError("Annotation is required", func=func, param=p_name)

            hint: type = hints[p_name]

            if get_origin(hint) in (UnionType, Union):
                type_args = get_args(hint)
                nullable = type(None) in type_args
                if not nullable or (nullable and len(type_args) >= 3):
                    raise InjectionError("Type unions are not allowed", func=func, param=p_name)
                hint = next(filter(lambda t: t is not NoneType, type_args))

            hint_t = Type(hint, lookup=vars_lookup)

            for resolver in [*additional_resolvers, *self._arg_resolvers]:
                options = ArgResolverOptions(
                    self,
                    func,
                    func_type_vars,
                    strategy,
                    p_name,
                    hint_t,
                    nullable,
                    default_set,
                    default,
                    immediate,
                    circular_guard,
                )
                if resolver.supports(options):
                    arg_name, arg_value = resolver.resolve(options)
                    f_args[arg_name] = arg_value
                    break

        if _args or _named_args:
            raise InjectionError(
                f"Expected {len(func_params)} arguments, {len(args) + len(named_args)} given",
                func=func,
            )

        return f_args

    def __hook_proxies__(self, instance: object) -> None:
        hooks: dict[str, Type[Any]] = {}
        cls = type(instance)
        cls_attrs: dict[str, Any] = dict(vars(cls))
        attr: InjectionHook[Any] | Any
        for name, attr in cls_attrs.items():
            if isinstance(attr, InjectionHook):
                delattr(cls, name)
                hooks[name] = attr.t
        instance_attrs: dict[str, Any] = dict(vars(instance))
        for name, attr in instance_attrs.items():
            if isinstance(attr, InjectionHook):
                delattr(instance, name)
                hooks[name] = attr.t
        for name, t in hooks.items():
            r_type = self._types[t.cls].get_type(t)
            setattr(cls, name, InjectionProxy(name, r_type, t))

    def _run_init_recursive(
        self,
        cls: type[InstanceT],
        instance: InstanceT,
        vars_lookup: TypeVarLookup[InstanceT] | None,
        circular_guard: OrderedSet[Any] | None,
    ) -> None:
        for base in cls.__bases__:
            self._run_init_recursive(base, instance, vars_lookup, circular_guard)
        for _, attr in vars(cls).items():
            if meta.has(attr, InitMethodMeta):
                self.call(attr, args=[instance], vars_lookup=vars_lookup, circular_guard=circular_guard)

    def _run_init_methods(
        self,
        r_type: "RegisteredType[InstanceT]",
        instance: InstanceT,
        vars_lookup: TypeVarLookup[InstanceT] | None,
        circular_guard: OrderedSet[Any],
    ):
        for method in r_type.before_init:
            self.call(method, args=[instance], circular_guard=circular_guard)
        self._run_init_recursive(r_type.t.cls, instance, vars_lookup, circular_guard)
        for method in r_type.after_init:
            self.call(method, args=[instance], circular_guard=circular_guard)

    def __instanciate__(
        self,
        r_type: "RegisteredType[InstanceT]",
        t: Type[InstanceT],
        circular_guard: OrderedSet[Any],
    ) -> InstanceT:
        vars_lookup = TypeVarLookup(t)
        func_args = self._resolve_args(
            r_type.t.cls,
            t.vars,
            r_type.strategy,  # type: ignore
            vars_lookup,
            False,
            circular_guard,
            r_type.args,
            r_type.named_args,
            [],
        )
        instance = r_type.t.cls(**func_args)
        self.__hook_proxies__(instance)
        meta.set(instance, self, cls=Injection)
        meta.set(instance, GenericMeta(t.vars))
        self._run_init_methods(r_type, instance, vars_lookup, circular_guard)
        self.__set_instance__(r_type, instance)
        return instance

    def is_registered(self, cls: type[Any] | Type[Any]) -> bool:
        if not isinstance(cls, Type):
            t = Type(cls)
        else:
            t = cls
        return t.cls in self._types and self._types[t.cls].is_registered(t)

    def call(
        self,
        func: Callable[..., FuncT],
        *,
        args: list[Any] | None = None,
        named_args: dict[str, Any] | None = None,
        vars_lookup: TypeVarLookup[Any] | None = None,
        additional_resolvers: list[ArgumentResolver] | None = None,
        circular_guard: OrderedSet[Any] | None = None,
    ) -> FuncT:
        func_args = self._resolve_args(
            func,
            None,
            "singleton",
            vars_lookup,
            True,
            circular_guard or OrderedSet(),
            args or [],
            named_args or {},
            additional_resolvers or [],
        )
        return func(**func_args)

    def instanciate(
        self,
        cls: type[InstanceT],
        *,
        args: list[Any] | None = None,
        named_args: dict[str, Any] | None = None,
        additional_resolvers: list[ArgumentResolver] | None = None,
    ) -> InstanceT:
        t = Type(cls)
        init_args = self._resolve_args(
            cls,
            t.vars,
            "immediate",
            None,
            True,
            OrderedSet(),
            args or [],
            named_args or {},
            additional_resolvers or [],
        )
        instance = cls(**init_args)
        self.__hook_proxies__(instance)
        meta.set(instance, self, cls=Injection)
        meta.set(instance, GenericMeta(t.vars))
        self._run_init_recursive(cls, instance, None, None)
        return instance

    def _add_type_instance(
        self,
        super_t: Type[InstanceT],
        t: Type[InstanceT],
        match_all: bool,
        strategy: InjectionStrategy,
        args: list[Any],
        named_args: dict[str, Any],
        before_init: list[Callable[[Any], None]],
        after_init: list[Callable[[Any], None]],
        instance: InstanceT | None,
        *,
        safe: bool = False,
    ) -> None:
        if super_t.cls not in self._types:
            self._types[super_t.cls] = RegisteredTypeBag(super_t)
        type_bag = self._types[super_t.cls]
        r_type: RegisteredType[InstanceT] | None = None
        if match_all:
            if not safe or not type_bag.has_match_all():
                r_type = type_bag.set_match_all(t, strategy, args, named_args, before_init, after_init)
        else:
            if not safe or not type_bag.has_type(super_t):
                r_type = type_bag.add_type(super_t, t, strategy, args, named_args, before_init, after_init)
        if instance is not None:
            if r_type is None:
                r_type = self._types[t.cls].get_type(t)
            if not safe or not self.__has_instance__(r_type):
                self.__set_instance__(r_type, instance)

    @overload
    def add(
        self,
        cls: type[InstanceT],
        strategy: InjectionStrategy,
        args: list[Any] | None = None,
        named_args: dict[str, Any] | None = None,
        instance: InstanceT | None = None,
        before_init: list[Callable[[InstanceT], None]] | None = None,
        after_init: list[Callable[[InstanceT], None]] | None = None,
        match_all: bool = False,
        super_cls: type[InstanceT] | None = None,
        *,
        instanciate: Literal[True],
    ) -> InstanceT:
        pass

    @overload
    def add(
        self,
        cls: type[InstanceT],
        strategy: InjectionStrategy,
        args: list[Any] | None = None,
        named_args: dict[str, Any] | None = None,
        instance: InstanceT | None = None,
        before_init: list[Callable[[InstanceT], None]] | None = None,
        after_init: list[Callable[[InstanceT], None]] | None = None,
        match_all: bool = False,
        super_cls: type[InstanceT] | None = None,
    ) -> None:
        pass

    def add(
        self,
        cls: type[InstanceT],
        strategy: InjectionStrategy,
        args: list[Any] | None = None,
        named_args: dict[str, Any] | None = None,
        instance: InstanceT | None = None,
        before_init: list[Callable[[InstanceT], None]] | None = None,
        after_init: list[Callable[[InstanceT], None]] | None = None,
        match_all: bool = False,
        super_cls: type[InstanceT] | None = None,
        *,
        instanciate: bool = False,
    ) -> InstanceT | None:
        if super_cls is None:
            super_cls = cls
        t = Type(cls)
        if hasattr(t.cls, "__parameters__"):
            if TYPE_CHECKING:
                assert isinstance(t.cls, _GenericOrigin)
            if len(t.cls.__parameters__) != len(t.vars) and not match_all:
                raise InjectionError(
                    f"Type {t} requires {len(t.cls.__parameters__)} generic parameters and {len(t.vars)} were given"
                )
        super_t = Type(super_cls)
        if not issubclass(t.cls, super_t.cls):
            raise InjectionError(f"Type {t} does not inherit from type {super_t}")
        if instance is not None:
            if instanciate:
                raise InjectionError(
                    f"Cannot instanciate {t.cls} if an instance is provided",
                )
            if not isinstance(instance, t.cls):
                raise InjectionError(f"Object provided must an instance of type {t.cls}")
            if strategy not in self._ADD_INSTANCE_STRATEGIES:
                formatted_strategies = _format_list(self._ADD_INSTANCE_STRATEGIES, final_sep=" or ")
                raise InjectionError(
                    f"Injection strategy for {t.cls} must be {formatted_strategies} if an instance is provided"
                )
        self._add_type_instance(
            super_t,
            t,
            match_all,
            strategy,
            args or [],
            named_args or {},
            before_init or [],
            after_init or [],
            instance,
        )
        if instanciate:
            return self.require(t.cls)  # type: ignore
        return None

    def require(self, cls: type[InstanceT]) -> InstanceT:
        t = Type(cls)
        if not self.is_registered(t):
            raise InjectionError(f"Type {t} is not a registered type in the injection system")
        r_type = self._types[t.cls].get_type(t)
        if r_type.strategy not in self._REQUIREABLE_STRATEGIES:
            formatted_strategies = _format_list(self._REQUIREABLE_STRATEGIES, final_sep=" or ")
            raise InjectionError(
                f"Injection strategy for {t} must be {formatted_strategies} to be required in this context"
            )
        if self.__has_instance__(r_type):
            return self.__get_instance__(r_type)
        return self.__instanciate__(r_type, t, OrderedSet())

    def get_scoped_session(self) -> "ScopedInjection":
        return ScopedInjection(self._cache, self._global_ctx, InjectionContext(), self._types)


class ScopedInjection(Injection):
    __slots__ = "_scoped_ctx"

    _ADD_INSTANCE_STRATEGIES = ("singleton", "scoped")
    _REQUIREABLE_STRATEGIES = ("singleton", "scoped", "transcient")

    def __init__(
        self,
        cache: Cache,
        global_ctx: InjectionContext,
        scoped_ctx: InjectionContext,
        types: "dict[type[Any], RegisteredTypeBag[Any]]",
    ) -> None:
        self._scoped_ctx = scoped_ctx
        super().__init__(cache, global_ctx, types)
        self._scoped_ctx.set_instance(Type(Injection), self)

    @override
    def __has_instance__(self, r_type: "RegisteredType[Any]") -> bool:
        return (r_type.strategy == "scoped" and self._scoped_ctx.has_instance(r_type.t)) or (
            r_type.strategy == "singleton" and self._global_ctx.has_instance(r_type.t)
        )

    @override
    def __get_instance__(self, r_type: "RegisteredType[InstanceT]") -> InstanceT:
        if self._scoped_ctx.has_instance(r_type.t):
            return self._scoped_ctx.get_instance(r_type.t)
        return self._global_ctx.get_instance(r_type.t)

    @override
    def __set_instance__(self, r_type: "RegisteredType[InstanceT]", instance: InstanceT) -> None:
        strategy = r_type.strategy
        if strategy == "scoped":
            self._scoped_ctx.set_instance(r_type.t, instance)
        if strategy == "singleton":
            self._global_ctx.set_instance(r_type.t, instance)

    @override
    def _pickup_resolvers(self, cache: Cache) -> "list[ArgumentResolver]":
        resolver_scoped: dict[type, bool] = {}
        resolver_priority: dict[type, int] = {}
        resolver_types: list[type[ArgumentResolver]] = []

        for t in cache.get(ArgumentResolver, hint=type[ArgumentResolver], raises=False):
            _meta = meta.get(t, ArgResolverMeta)
            resolver_priority[t] = _meta.priority
            resolver_scoped[t] = _meta.scoped
            resolver_types.append(t)
        resolver_types = sorted(resolver_types, key=lambda t: resolver_priority[t])

        resolvers: list[ArgumentResolver] = []
        for cls in resolver_types:
            resolvers.append(self.instanciate(cls, additional_resolvers=[DefaultArgResolver()]))

        return [*resolvers, DefaultArgResolver()]

    @override
    def call(
        self,
        func: Callable[..., FuncT],
        *,
        args: list[Any] | None = None,
        named_args: dict[str, Any] | None = None,
        vars_lookup: TypeVarLookup[Any] | None = None,
        additional_resolvers: list[ArgumentResolver] | None = None,
        circular_guard: OrderedSet[Any] | None = None,
    ) -> FuncT:
        func_args = self._resolve_args(
            func,
            None,
            "scoped",
            vars_lookup,
            True,
            circular_guard or OrderedSet(),
            args or [],
            named_args or {},
            additional_resolvers or [],
        )
        return func(**func_args)


class _GenericOrigin(Protocol):
    __parameters__: tuple[TypeVar, ...]


def _format_list(collection: Collection[Any], *, sep: str = ", ", final_sep: str | None = None) -> str:
    """TODO: move this into StringUtils and rework import flow"""
    formatted: list[str] = []
    cnt = len(collection)
    for i, e in enumerate(collection):
        formatted.append(str(e))
        if i != cnt - 1:
            if i == cnt - 2 and final_sep:
                formatted.append(final_sep)
            else:
                formatted.append(sep)
    return "".join(formatted)



from typing import Any, Generic, Iterator, TypeVar

from bolinette.core.types import Type

T = TypeVar("T")


class TypeVarLookup(Generic[T]):
    __slots__ = ("t", "_lookup")

    def __init__(self, t: Type[T]) -> None:
        self.t = t
        self._lookup = TypeVarLookup.get_lookup(t)

    @property
    def empty(self) -> bool:
        return self._lookup is None or len(self._lookup) == 0

    def __getitem__(self, key: TypeVar) -> type[Any]:
        if self._lookup is None or key not in self._lookup:
            raise KeyError(key)
        return self._lookup[key]

    def __contains__(self, key: TypeVar) -> bool:
        return self._lookup is not None and key in self._lookup

    def __iter__(self) -> Iterator[tuple[TypeVar, type]]:
        if self._lookup is None:
            return iter(())
        return ((t, v) for t, v in self._lookup.items())

    @staticmethod
    def get_lookup(t: Type[Any]) -> dict[TypeVar, type[Any]] | None:
        if not hasattr(t.cls, "__parameters__"):
            return None
        return {n: t.vars[i] for i, n in enumerate(t.cls.__parameters__)}



from setuptools import find_namespace_packages, setup

from bolinette.core import __version__
from bolinette.core.utils import FileUtils, PathUtils

paths = PathUtils()
files = FileUtils(paths)


def project_packages(module: str) -> list[str]:
    return [m for m in find_namespace_packages() if m.startswith(module)]


setup(
    name="bolinette",
    packages=project_packages("bolinette.core"),
    include_package_data=True,
    version=__version__,
    license="MIT",
    description="The Bolinette core package, an inversion of control framework",
    long_description=files.read_file(paths.root_path("README.md")),
    long_description_content_type="text/markdown",
    author="Pierre Chat",
    author_email="pierrechat@outlook.com",
    url="https://github.com/bolinette/bolinette",
    keywords=["Bolinette", "Framework", "IoC", "Dependency Injection"],
    install_requires=files.read_requirements(paths.root_path(), name="requirements.core.txt"),
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Build Tools",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.11",
    ],
    setup_requires=["wheel"],
    entry_points={
        "console_scripts": ["blnt=bolinette.core.__main__:main"],
    },
)



from bolinette.core import Cache
from bolinette.core.exceptions import InitError


class Extension:
    def __init__(self, cache: Cache, dependencies: "list[Extension] | None" = None) -> None:
        self.cache = cache
        self.dependencies = dependencies or []

    @staticmethod
    def sort_extensions(extensions: "list[Extension]") -> "list[Extension]":
        dependencies: dict[Extension, set[Extension]] = {}
        for extension in extensions:
            dependencies[extension] = set(extension.dependencies)
        sorted_extensions: list[Extension] = []
        while len(sorted_extensions) != len(extensions):
            no_dependencies: list[Extension] = []
            for extension, deps in dependencies.items():
                if not deps:
                    no_dependencies.append(extension)
            if len(no_dependencies) == 0:
                raise InitError("A circular dependency was detected in the loaded extensions")
            for extension in no_dependencies:
                sorted_extensions.append(extension)
                del dependencies[extension]
                for other_extension in dependencies:
                    if extension in dependencies[other_extension]:
                        dependencies[other_extension].remove(extension)
        return sorted_extensions

    @staticmethod
    def merge_caches(extensions: "list[Extension]") -> Cache:
        cache = Cache()
        for extension in extensions:
            cache |= extension.cache
        return cache


core_ext = Extension(Cache())



from collections.abc import Iterator
from typing import Any, Callable

from typing_extensions import override


class BolinetteError(Exception):
    def __init__(self, message: str) -> None:
        Exception.__init__(self, message)
        self.message = message


class InitError(BolinetteError):
    def __init__(self, message: str) -> None:
        BolinetteError.__init__(self, message)
        self.message = message

    @override
    def __str__(self) -> str:
        return self.message


class InternalError(BolinetteError):
    pass


class ParameterError:
    def __init__(self, **params: str) -> None:
        self._error_params = params

    def _format_params(self, message: str, **values: Any) -> str:
        f_strings: list[str] = []
        for param, f_string in self._error_params.items():
            if param in values and values[param]:
                f_strings.append(f_string.replace("{}", str(values[param])))
        return ", ".join(f_strings + [message])


class ErrorCollection(Exception):
    def __init__(self, errors: list[BolinetteError] | None = None) -> None:
        self._errors: list[BolinetteError] = errors or []
        Exception.__init__(self, "\n".join(e.message for e in self._errors))

    def append(self, error: BolinetteError) -> None:
        self._errors.append(error)

    def __bool__(self) -> bool:
        return any(self._errors)

    def __iter__(self) -> Iterator[BolinetteError]:
        return iter(self._errors)

    @override
    def __str__(self) -> str:
        return str(self._errors)

    @override
    def __repr__(self) -> str:
        return f'<BolinetteErrors [{",".join([repr(err) for err in self._errors])}]>'


class TypingError(BolinetteError, ParameterError):
    def __init__(
        self,
        message: str,
        *,
        cls: str | None = None,
    ) -> None:
        ParameterError.__init__(self, cls="Type {}")
        BolinetteError.__init__(self, self._format_params(message, cls=cls))


class InjectionError(BolinetteError, ParameterError):
    def __init__(
        self,
        message: str,
        *,
        cls: type[Any] | None = None,
        func: Callable[..., Any] | None = None,
        param: str | None = None,
    ) -> None:
        ParameterError.__init__(self, cls="Type {}", func="Callable {}", param="Parameter '{}'")
        BolinetteError.__init__(
            self,
            self._format_params(
                message,
                cls=cls,
                func=func.__qualname__ if func is not None else None,
                param=param,
            ),
        )


class EnvironmentError(BolinetteError):
    pass


class ExpressionError(BolinetteError):
    def __init__(self, message: str) -> None:
        super().__init__(f"Expression error: {message}")


class MappingError(BolinetteError, ParameterError):
    def __init__(
        self,
        message: str,
        *,
        src: str | None = None,
        dest: str | None = None,
    ) -> None:
        ParameterError.__init__(self, dest="Destination path '{}'", src="From source path '{}'")
        BolinetteError.__init__(self, self._format_params(message, dest=dest, src=src))


class InitMappingError(InitError):
    pass



from bolinette.core.mapping.profiles import mapping as mapping, Profile as Profile
from bolinette.core.mapping.mapper import Mapper as Mapper, type_mapper as type_mapper, MappingRunner as MappingRunner



import os


class PathUtils:
    def __init__(self) -> None:
        self._cwd = self.cwd()

    def root_path(self, *path: str) -> str:
        return self.join(self._cwd, *path)

    def instance_path(self, *path: str) -> str:
        return self.root_path("instance", *path)

    def env_path(self, *path: str) -> str:
        return self.root_path("env", *path)

    def static_path(self, *path: str) -> str:
        return self.root_path("static", *path)

    def templates_path(self, *path: str) -> str:
        return self.root_path("templates", *path)

    @staticmethod
    def cwd() -> str:
        return os.getcwd()

    @staticmethod
    def mkdir(path: str) -> None:
        os.makedirs(path)

    @staticmethod
    def exists(path: str) -> bool:
        return os.path.exists(path)

    @staticmethod
    def rename(path: str, new_path: str) -> None:
        os.rename(path, new_path)

    @staticmethod
    def join(*args: str) -> str:
        return os.path.join(*args)

    @staticmethod
    def dirname(path: str) -> str:
        return os.path.dirname(os.path.realpath(path))

    @staticmethod
    def split(path: str) -> tuple[str, str]:
        return os.path.split(path)

    @staticmethod
    def rm(path: str, *, recursive: bool = False) -> None:
        if recursive:
            for root, dirs, files in os.walk(path, topdown=False):
                for name in files:
                    os.remove(os.path.join(root, name))
                for name in dirs:
                    os.rmdir(os.path.join(root, name))
            os.rmdir(path)
        else:
            os.remove(path)
